"""Provides a visual panel (L{qubx.faces.Face}) to idealize single-molecule data.

Uses math from U{qubx_single_channel<http://www.qub.buffalo.edu/src/qub-express/api/qubx_single_channel>}

/* Copyright 2008-2014 Research Foundation State University of New York   */

/* This file is part of QUB Express.                                      */

/* QUB Express is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Express is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Express program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

"""

import itertools
import gc
import os
import sys
import time
import traceback
import gobject

from qubx.data_types import *
from qubx.util_types import *
from qubx.faces import *
from qubx.GTK import *
import qubx.fast.model
import qubx.idealize
import qubx.notebook
import qubx.optimize
import qubx.optimizeGTK
import qubx.pyenv
import qubx.settings
import qubx.tree
from qubx.model import *
from qubx.settings import Property, Propertied
from qubx.task import *
import qubx_idealize
from qubx_idealize import QUBX_IDLZ_METHODS

import scipy.optimize
from numpy import *

import ctypes
from ctypes import c_int, c_float, c_double, c_void_p, c_char_p, Structure, POINTER, CFUNCTYPE, byref, sizeof, py_object, pointer, cast
pybuf = ctypes.pythonapi.PyBuffer_FromReadWriteMemory
pybuf.restype = py_object


MAXITER = 300
MAX_POINTS_AT_ONCE = (1 << 24)

def get_data(segs, receiver, last_data=None):
    dataview = qubx.pyenv.env.globals['QubX'].Data.view
    if not dataview:
        gobject.idle_add(receiver, None)
        return
    
    stimulus = qubx.pyenv.env.globals['QubX'].Modeling.Stimulus
    stim_names = stimulus.active_names
    if last_data:
        data = last_data
        data.reset(dataview.file.sampling, len(stim_names))
    else:
        data = qubx_idealize.Data(dataview.file.sampling, len(stim_names)) # , data_cap_samples)
        
    sampling_sec = dataview.file.sampling
    if 'Voltage' in stim_names:
        data.iVoltage = stim_names.index('Voltage') + 1
    else:
        data.iVoltage = 0
    data.signal_names = ['Current'] + stim_names
    data.ff_ll_counts = []

        # segment is divided in chunks, alternating included/excluded (data / no data)
        # each segment, we request idl of each signal, each chunk
        # this could involve re-idealization, indefinite time, so result comes via callback
        # we provide generator instance as callback, repeatedly, to process segments in a loop
        
        # gobject thread:                               stimulus robot:
        #   rcv_stims.next()
        #      repeat:
        #         stimulus.call_with_stimulus_idl
        #         yield...
        #   [get(self, ...) returns]
        #                                                  idealize stim if needed
        #                                                  idle_add(rcv_stims.send, ...)
        #         ...process chunks
        #      yield None

    rcv_stims = None # will have instance of receive_stims()
    def rcv_hack(*args): # can send only one argument into a generator
        rcv_stims.send(args)
    def receive_stims():
        iseg = -1
        for seg in segs:
            data.file = seg.file
            data.iSignal = seg.signal
            stimulus.call_with_stimulus_idl(rcv_hack, [(chunk.f, chunk.l) for chunk in seg.chunks])
            stim_names, stimidl = (yield)
            chunkidl = [ Anon(dwellCounts=[], classeses=[], durationses=[], ampses=[]) for chunk in seg.chunks]
            for segs_flcda in stimidl: # for each signal
                for ichunk, flcda in enumerate(segs_flcda): # each chunk
                    ff, ll, cc, dd, aa  = flcda
                    chunkidl[ichunk].dwellCounts.append(len(cc))
                    chunkidl[ichunk].classeses.append(cc)
                    chunkidl[ichunk].durationses.append(dd)
                    chunkidl[ichunk].ampses.append(aa)
            
            iseg += 1
            data.append_segment(seg.f)
            for ichunk, chunk in enumerate(seg.chunks):
                Nd = chunk.n
                yy = None
                if chunk.included:
                    chunkoff = chunk.f - seg.f
                    yy = zeros(shape=(Nd,), dtype='float32')
                    ii = 0
                    for ch in generate_chunk_samples([chunk]):
                        yy[ii:ii+ch.n] = ch.samples
                        ii += ch.n
                    ff = arange(seg.f+chunkoff, seg.f+chunkoff+Nd, dtype='int32')
                    ll = ff + 1
                    data.ff_ll_counts.append( (ff, ll, Nd) )
                data.append_chunk(iseg, Nd, yy,
                                  chunkidl[ichunk].dwellCounts, chunkidl[ichunk].classeses,
                                  chunkidl[ichunk].durationses, chunkidl[ichunk].ampses)
        gobject.idle_add(receiver, data)
        yield None

    rcv_stims = receive_stims()
    rcv_stims.next()


@Propertied(Property("usePeq", True, "True to start each segment at equilibrium; False to use States:Pr"),
            Property("Gleak", 0.0, "Leakage conductance (pS)"),
            Property("Vleak", 0.0, "Leak voltage (mV)"),
            Property("resolution", 0.05, "Gaussian sampling (pA)"),
            Property("reest_amps", True, "True to recompute amplitudes each iteration"),
            Property("reest_rates", False, "True to recompute transition probabilities each iteration"),
            Property("baseline", False, "True to track baseline with Kalman filter"),
            Property("baseline_std", 1e-3, "Kalman process standard deviation"),
            Property("baseline_nodes", True, "True to add baseline nodes, if baseline"),
            Property("baseline_nodes_delta", 0.05, "Minimum y-distance between nodes")
            )
class Idealizers(gtk.VBox):
    __explore_featured = ['global_name', 'robot', 'shared_controls', 'method', 'qubx_finalize', 'idealize']
    def __init__(self):
        gtk.VBox.__init__(self)
        self.global_name = 'QubX.Modeling.Idealize.methods["Threshold"]'
        self.__ref = Reffer()
        self.propertied_connect_settings('qubx_single.idealize')
        self.__running = False
        self.set_size_request(200, 50)
        self.QubX = qubx.pyenv.env.globals['QubX']
        self.QubX.OnQuit += self.__ref(self.qubx_finalize)
        self.robot = Robot('Idealize', self.__ref(lambda: Tasks.add_task(self.robot)),
                                       self.__ref(lambda: Tasks.remove_task(self.robot)))
        self.robot.OnException += self.__ref(self.__onException)
        self.robot.OnInterrupt += self.__ref(self.__onInterrupt)
        
        llstuf = pack_item(gtk.HBox(), self)
        col = pack_item(gtk.VBox(), llstuf)
        self.chkPeq = pack_check('start at equilibrium p', col)
        self.chkPeq.set_tooltip_text('otherwise use States:Pr')
        self.propertied_connect_check('usePeq', self.chkPeq)
        
        pack_space(llstuf, x=20)
        col = pack_item(gtk.VBox(), llstuf)
        h = pack_item(gtk.HBox(), col)
        pack_label('Gleak:', h)
        self.txtGleak = pack_item(NumEntry(self.Gleak, acceptFloat, '%.3g', width_chars=6), h, at_end=True)
        self.propertied_connect_NumEntry('Gleak', self.txtGleak)
        
        pack_space(llstuf, expand=True)
        col = pack_item(gtk.VBox(), llstuf)
        h = pack_item(gtk.HBox(), col)
        pack_label('Vleak:', h)
        self.txtVleak = pack_item(NumEntry(self.Vleak, acceptFloat, '%.3g', width_chars=6), h, at_end=True)
        self.propertied_connect_NumEntry("Vleak", self.txtVleak)

        #pack_space(llstuf, expand=True)
        #hint = 'delta y for integrating Gaussian amp probability'
        #col = pack_item(gtk.VBox(), llstuf)
        #row = pack_item(gtk.HBox(), col)
        #lbl = pack_label('y resolution:', row)
        #lbl.set_tooltip_text(hint)
        #self.txtResolution = pack_item(NumEntry(self.resolution, acceptFloatGreaterThan(0.0), '%.2g', width_chars=5), row)
        #self.txtResolution.set_tooltip_text(hint)
        #self.propertied_connect_NumEntry("resolution", self.txtResolution)
        
        optstuf = pack_item(gtk.HBox(), self, expand=True)
        self.txtStatus = gtk.TextView()
        self.outStatus = TextViewAppender(self.txtStatus)
        self.txtStatus.set_editable(False)
        SetFixedWidth(self.txtStatus)
        pack_scrolled(self.txtStatus, optstuf, expand=True)
        self.outStatus.write("""Based on "Restoration of single-channel currents using the segmental k-means method based on hidden Markov modeling," Qin F, Biophys J. 2004 Mar; 86(3):1488-501.

Before idealizing, make sure to "Grab" the amplitude distributions)
(right-click a state in the model).
""")
        col = pack_item(gtk.VBox(), optstuf)
        self.chkReestAmps = pack_check("re-est amps", col)
        self.propertied_connect_check("reest_amps", self.chkReestAmps)
        self.chkReestRates = pack_check("re-est rates", col)
        self.propertied_connect_check("reest_rates", self.chkReestRates)
        row = pack_item(gtk.HBox(), col)
        self.chkBaseline = pack_check("baseline std:", row)
        self.propertied_connect_check("baseline", self.chkBaseline)
        self.txtBaseline = pack_item(NumEntry(self.baseline_std, acceptFloatGreaterThan(0.0), '%.3g', width_chars=6), row, expand=True)
        self.propertied_connect_NumEntry("baseline_std", self.txtBaseline)
        row = pack_item(gtk.HBox(), col)
        self.chkBaselineNodes = pack_check("add nodes, delta:", row)
        self.propertied_connect_check("baseline_nodes", self.chkBaselineNodes)
        self.chkBaselineNodes.set_tooltip_text('To see and edit baseline nodes, choose the Baseline tool in the Data panel')
        self.txtBaselineDelta = pack_item(NumEntry(self.baseline_nodes_delta, acceptFloatGreaterThan(0.0), '%.3g', width_chars=6), row, expand=True)
        self.propertied_connect_NumEntry('baseline_nodes_delta', self.txtBaselineDelta)
        self.txtBaselineDelta.set_tooltip_text('Minimum vertical distance between nodes')
        self.shared_controls = pack_item(gtk.HBox(), col, at_end=True)
        
        self.__method = "Threshold"
    def set_method(self, x):
        self.__method = x
        self.__sensitize()
    method = property(lambda self: self.__method, lambda self, x: self.set_method(x))
    def propertied_set(self, value, name):
        super(Idealizers, self).propertied_set(value, name)
        self.__sensitize()
    def __sensitize(self):
        thresh = self.method == 'Threshold'
        self.chkReestAmps.set_sensitive(not thresh)
        self.chkReestRates.set_sensitive(not thresh)
        self.chkBaseline.set_sensitive(not thresh)
        self.txtBaseline.set_sensitive((not thresh) and self.baseline)
        self.chkBaselineNodes.set_sensitive((not thresh) and self.baseline)
        self.txtBaselineDelta.set_sensitive((not thresh) and self.baseline and self.baseline_nodes)
    def qubx_finalize(self):
        self.robot.stop()
        #self.QubX.OnQuit -= self.__ref(self.qubx_finalize)
    def __onException(self, robot, typ, val, trace):
        traceback.print_exception(typ, val, trace, file=self.outStatus)
    def __onSetDwells(self, dwells_per_seg, nodes_per_seg):
        for ff, ll, cc, aa, ss in dwells_per_seg:
            self.idl_updater.set_dwells(len(ff), ff, ll, cc, aa, ss, event=True)
        if nodes_per_seg:
            for ff, ll, cc, aa, ss in dwells_per_seg:
                if len(ff):
                    self.datafile.baseline[self.iSignal].clear_nodes(ff[0], ll[-1])
            for nodes in nodes_per_seg:
                self.__set_nodes(*nodes)
    def __onDoneIdl(self, ll, segs):
        self.outStatus.write('Done.\n')
        if ll:
            self.outStatus.write("LL: %f\n" % ll)
    def __set_nodes(self, points, values):
        self.datafile.baseline[self.iSignal].add_nodes(points, values)
    def __onInterrupt(self, robot, cancel):
        cancel() # the flag is good enough, while the KeyboardInterrupt can bewilder it
        if isinstance(self.robot.model, qubx.fast.model.Model):
            self.robot.model.obj[0].stop_flag = 1
        elif self.robot.model:
            self.robot.model[0].stop_flag = 1
    def __onProgressTimer(self):
        if self.__running:
            curr_progress = self.robot.model.progress if isinstance(self.robot.model, qubx.fast.model.Model) else self.robot.model[0].progress
            self.robot.progress = self.robot.progress_before + int(round(self.robot.progress_segs * 1e-2 * curr_progress))
            return True
    def robot_setup(self, mtree, Peq, Gleak, Vleak, segs):
        self.robot.mtree = mtree
        self.robot.Peq = Peq
        self.robot.Gleak = Gleak
        self.robot.Vleak = Vleak
        self.robot.data = self.robot.gui_call_recv(get_data, segs, self.robot)[0]
        self.robot.signal_names = self.robot.data.signal_names
        self.robot_setup_model(self.robot.data)
        self.__running = True
        gobject.timeout_add(500, self.__onProgressTimer)
    def robot_setup_model(self, data):
        model = QubModel()
        if self.robot.mtree:
            model.from_tree(self.robot.mtree)
        try:
            self.robot.single_model.dispose() # work around callback-related memory leak
        except:
            pass
        self.robot.single_model = ModelToFast(model, data.signal_names, self.robot_on_status)
        self.robot.single_model.usePeq = self.robot.Peq
        self.robot.single_model.iVoltage = data.iVoltage
        self.robot.single_model.Gleak = self.robot.Gleak ### TODO: move into model?
        self.robot.single_model.Vleak = self.robot.Vleak ###
        if model.channelCount > 16:
            self.robot_on_status("Warning: Channel count > 16 not supported; setting it to 16.")
            self.robot.single_model.Nchannel = 16

        self.robot.multi_model = qubx.fast.model.MultiModel(self.robot.single_model)
        self.robot.model = self.robot.multi_model.multi_model
        self.robot.K0, self.robot.K1, self.robot.L, self.robot.V = ModelToRateMatrices(model, data.signal_names)
        self.robot.multi_model.update()

        stimclasses = self.robot.data.get_stimclasses()
        Nstim = self.robot.data.Nstim
        Nsig = self.robot.data.Nsig
        sc_frac = self.robot.data.get_stimclass_frac()
        self.robot.ampm = qubx.fast.model.StimAmps(self.robot.model, Nstim, Nsig, stimclasses, sc_frac)
        if self.robot.method != 'Threshold': # skip for half-amp
            self.robot.ratesm = qubx.fast.model.StimRates(self.robot.model, Nstim, Nsig, stimclasses, sc_frac,
                                                          self.robot.data.sampling)
            A, B = model.constraints_kin.get_matrices(self.robot.K0, self.robot.K1, self.robot.L, self.robot.V, auto=False)
            if A != None:
                self.robot.ratesm.cns.Ain[:A.shape[0],:A.shape[1]] = A
                self.robot.ratesm.cns.Bin[:B.shape[0]] = B.flatten()
                self.robot.ratesm.cns.Ncns = B.shape[0]
            else:
                self.robot.ratesm.cns.Ncns = 0
    def idealize(self, segments, model, on_finish):
        QubX = qubx.global_namespace.QubX
        if ((QubX.DataSource.source == qubx.data_types.DATASOURCE_LIST) or
            (      (QubX.DataSource.source == qubx.data_types.DATASOURCE_SCREEN) and
                   (len(segments) == 1) and
                   QubX.Data.file.list and (0 <= QubX.Data.file.list.find_selection(segments[0].f, segments[0].l)))):
            self.__out_list = QubX.Data.file.list
            self.__out_list_ix = QubX.Data.file.list.find_selection(segments[0].f, segments[0].l) if (QubX.DataSource.source == qubx.data_types.DATASOURCE_SCREEN) else 0
        else:
            self.__out_list = None
        self.idl_updater = segments[0].file.update_idl(segments[0].signal) if segments else None
        if model.states and segments:
            self.robot.do(self.robot_idealize, segments, model, on_finish)
        else:
            gobject.idle_add(on_finish)
    def robot_idealize(self, segments, model, on_finish):
        dwells = None
        with self.robot.main_hold:
            self.datafile = segments[0].file
            self.__model = model
            self.iSignal = segments[0].signal
            mtree = model.as_tree()
            Peq = self.usePeq
            Gleak = self.Gleak
            Vleak = self.Vleak
            amps = mtree['Amps'].data # doesn't honor IeqFv
            self.robot.method = self.method
            self.robot.resolution = 10**int(round(log10((max(amps) - min(amps)) * 1e-2))) # self.resolution
            self.robot.reest_amps = self.reest_amps
            self.robot.reest_rates = self.reest_rates
            self.robot.track_baseline = self.baseline
            self.robot.baseline_std = self.baseline_std
            self.robot.baseline_nodes = self.baseline_nodes
            self.robot.baseline_nodes_delta = self.baseline_nodes_delta
            self.robot.segs = self.QubX.DataSource.get_segmentation(baseline_nodes=(not self.robot.track_baseline)) # use raw data when tracking baseline
        segs_lst = break_segs(segments, MAX_POINTS_AT_ONCE)
        try:
            self.robot.progress_before = 0
            self.robot.progress_segs = segs_lst and (100.0 / len(segs_lst)) or 100.0
            ll = 0.0
            for segs in segs_lst:
                self.robot_setup(mtree, Peq, Gleak, Vleak, segs)
                if self.robot.method == 'Threshold':
                    ll += qubx_idealize.halfamp(self.robot.model, self.robot.data, self.robot.ampm)
                else:
                    ll += qubx_idealize.skm(self.robot.model, self.robot.data, self.robot.ampm, self.robot.ratesm,
                                            self.robot.resolution, (self.robot.method == 'SKM') and 2 or 1,
                                            self.robot.reest_amps, self.robot.reest_rates,
                                            self.robot.track_baseline, self.robot.baseline_std)
                dwells = [self.robot_renumber_dwell_classes(*self.robot.data.get_dwells(s)) for s in xrange(len(segs))]
                if self.robot.track_baseline and self.robot.baseline_nodes:
                    nodes = [resample_baseline(self.robot.data.get_baseline(s), self.robot.baseline_nodes_delta, seg.f) for s, seg in enumerate(segs)]
                else:
                    nodes = None
                gobject.idle_add(self.__onSetDwells, dwells, nodes)
                self.robot.progress_before += self.robot.progress_segs
        except:
            self.robot_on_status(traceback.format_exc())
            ll = 0.0
        self.__running = False
        gobject.idle_add(self.__after_set_dwells, ll)
        gobject.idle_add(on_finish)
    def __after_set_dwells(self, ll):
        qubx.notebook.Notebook.send(qubx.notebook.NbText("""Idealize: %s
    model: %s
    method: %s
    re-estimate: %s
    track baseline: %s
    resolution: %s
    LL: %f""" % (self.datafile.path,
                         self.__model.path,
                         self.robot.method,
                         "%s%s" % ((self.robot.reest_amps and "amps" or ""),
                                   (self.robot.reest_rates and "rates" or "")),
                         (self.robot.track_baseline and ("""yes
    baseline std: %.4g""" % self.robot.baseline_std) or "no"),
                         self.robot.resolution, ll)), auto_id='Idealize.Messages')
        if self.idl_updater:
            self.idl_updater.done()
        self.datafile = self.__model = self.__segs = self.__out_list = self.idl_updater = None
    def robot_on_status(self, msg):
        gobject.idle_add(self.outStatus.write, msg+"\n")
    def robot_renumber_dwell_classes(self, ff, ll, cc, aa, ss):
        if (self.robot.single_model.Nchannel == 1) and len(ff) and len(aa):
            class_order = self.robot.multi_model.class_order[:self.robot.multi_model.nclass_order]
            nc = max(class_order) + 1
            cc = numpy.array([class_order[c] for c in cc], dtype='int32')
            aa2 = numpy.zeros(shape=(nc,), dtype='float64')
            ss2 = numpy.zeros(shape=(nc,), dtype='float64')
            for ic, c in enumerate(class_order):
                if ic < len(aa):
                    aa2[c] = aa[ic]
                    ss2[c] = ss[ic]
            aa, ss = aa2, ss2
        return ff, ll, cc, aa, ss


# [.3, .5, .25, 1.2{.6,.6}, .3]
# [ [.3, .5], [.25], [.6{.6}], [.6{.6}], [.3] ]

def break_segs(segs, maxpoints):
    segs_lst = []
    segs_one = []
    sofar = 0
    for seg in segs:
        n = ((seg.chunks and sum(chunk.included and chunk.n or 0 for chunk in seg.chunks))
             or (seg.included and seg.n or 0))
        if (sofar + n) > maxpoints:
            if segs_one: # flush
                segs_lst.append(segs_one)
                segs_one = []
                sofar = 0
        if n > maxpoints:
            if seg.chunks:
                chunks = break_segs(seg.chunks, maxpoints)
                for chunklst in chunks:
                    segs_lst.append([Anon(seg, f=chunklst[0].f, l=chunklst[-1].l, n=(chunklst[-1].l - chunklst[0].f + 1),
                                               chunks=chunklst)])
            else:
                K = int(ceil(n*1.0/maxpoints))
                sub_n = n / K
                for k in xrange(K):
                    segs_lst.append([Anon(seg, f=seg.f+k*sub_n, l=min(seg.f+(k+1)*sub_n - 1, seg.l),
                                               n=min(sub_n, seg.l - (seg.f+k*sub_n) + 1))])
        else:
            segs_one.append(seg)
            sofar += n
    if segs_one: # flush
        segs_lst.append(segs_one)
    return segs_lst



# baseline as nodes
def resample_baseline(baseline, resolution, first):
    means, stds, firsts, lasts, closests = qubx.fast.data.adaptive_resample(baseline, resolution)
    closests += first
    return closests, means



# Gleak, Vleak hidden unless IeqFv? (to Models table?)
# how well does skm iterate?
